# -*- coding: utf-8 -*-

import os
from dataclasses import dataclass

from ysql import Dao, Sql
from ysql import Path, MetaDatabase
from ysql import Entity, Constraint


# ====================================================================================================================
# 项目gitee地址：https://gitee.com/darlingxyz/ysql（请于此查看项目详细信息，如README文件）
# 如果直接下载源码使用，请将ysql文件夹放置在根目录中，与venv目录同级，否则需要重新导入其中各文件的依赖项
# 通过pip安装时，可以在任何py文件中运行以下代码来生成此示例文件
# from ysql import get_example
# get_example()
# ====================================================================================================================
# 结构说明：
# Entity定义数据表
# Dao定义数据操作接口，提供表级控制
# Database集成Dao便于统一对外，提供库级控制


# ====================================================================================================================
# 数据表类的定义形式
@Entity  # 两个装饰器的位置先后可任意
@dataclass  # 显式的使用dataclass装饰器，可以在IDE中获取代码补全
class Student:  # 类名即表名
    picture: bytes
    name: str = Constraint.not_null, Constraint.unique  # 用逗号分个多个约束
    weight: float = 50, Constraint.not_null  # 若该字段存在默认值，同样以逗号分隔，但仅支持int、float、str、bytes类型的默认值；或者使用Constraint.default()实现约束
    student_id: int = Constraint.auto_primary_key


@Entity
@dataclass
class Score:
    score: float
    student_id: int = Constraint.auto_primary_key, \
                      Constraint.foreign_key(entity=Student,
                                             field='student_id',
                                             delete_link=Constraint.cascade,
                                             update_link=Constraint.set_default)  # 设置外键的具体约束


# ====================================================================================================================
# Dao类的定义形式
@Dao(entity=Student)  # 使用Dao装饰器，并传递对应的entity类
class DaoStudent:

    # 内置insert(entity)方法的返回值为刚插入记录的row id
    # 缺点是在IDE的自动补全无法提示该方法的存在
    # @Insert  # 如需重新定义插入方法，可使用sql装饰器并传递相应的sql语句，或者直接使用Insert装饰器，也可自动实现。
    # def insert(self, entity):
    #   pass
    # 如果使用了@Insert（包括内置的insert方法），被装饰方法会自动返回插入记录的主键

    @Sql("select * from student where student_id=? and name=?;")  # 对方法使用Sql装饰器，并传入sql语句即可
    def get_student(self, student_id, name):  # 参数名可任意，但顺序需要与sql语句中?的传值顺序一致
        pass  # 方法本身无需实现
        # 使用select的时候返回结果为list[named_tuple]，其中每条记录会自动解析为具名元组。


@Dao(entity=Score)  # 也可直接传入对应的entity
class DaoScore:

    # 涉及使用对应entity的表名时，可以使用__代替符，但相应的不允许sql语句在其他位置使用__字符。
    @Sql(f"select * from __ where student_id=?;")
    def get_score(self, student_id):
        pass  # 方法本身无需实现


# ====================================================================================================================
# 数据库类的定义形式
db_path = "db_folder/test_detailed.db"  # 此处是需要每次运行均删除数据库，故额外定义路径变量
@Path(db_path)  # 使用Path装饰器传递数据库的路径，可直接传路径字符串
class MyDatabase(MetaDatabase):  # 还需额外继承MetaDatabase类，原因是若采用装饰器实现，则无法获得自动补全的方法提示，查看MetaDatabase源码了解外部可调用的方法
    # 需将全部Dao类分别定义为类中静态变量，以便集成Dao到一个数据类中对外访问
    dao_student = DaoStudent()
    dao_score = DaoScore()


if __name__ == '__main__':
    # 删除已存在的数据库文件，方便每次重置
    if os.path.exists(db_path):
        os.remove(db_path)

    db = MyDatabase()
    db.connect()
    db.create_tables()

    db.dao_student.insert(Student(name="好家伙", picture=b'1235'))  # 插入一条student记录
    db.dao_score.insert(Score(score=59.5, student_id=1))  # 插入一条score记录
    db.commit()  # 提交改动

    result1 = db.dao_student.get_student(student_id=1, name='好家伙')  # 获取一条student记录
    result2 = db.dao_score.get_score(student_id=1)

    # 查询结果以列表形式返回
    print(f"查询结果1：{result1}\n查询结果2：{result2}")

    # 查询结果的单条记录为具名元组
    for result in result1:
        print(f"结果1中的姓名：{result.name}；图片：{result.picture}")

    db.disconnect()  # 关闭数据库连接



# -*- coding: utf-8 -*-

from ysql.entity import Constraint, get_constraint_from_property

# 约定的表名代替符，在sql语句中凡是涉及本表表名的，均可采取此代替符，内部可自动替换。
TABLE_SUBSTITUTE = '__'
DECORATOR_SQL_PARAMS = 'decorator_sql_params'


# ====================================================================================================================
# 装饰器方法
def Dao(entity):
    """dao类的统一装饰器"""

    def decorator(cls):
        # 新增cursor属性
        setattr(cls, "cursor", None)
        # 新增entity属性
        setattr(cls, "entity", entity)
        # 新增更新游标方法
        setattr(cls, update_cursor.__name__, update_cursor)
        # 新增生成表方法
        setattr(cls, get_sql_create_table.__name__, get_sql_create_table)
        # 新增插入方法
        setattr(cls, "insert", insert)
        return cls

    return decorator


def Sql(sql_command):
    """一般化执行sql语句的装饰器"""

    def decorator(func):  # noqa
        def wrapper(self, *args, **kwargs):

            # 如果包含 'decorator_sql_params' 参数，且传入的sql参数是可调用的
            if DECORATOR_SQL_PARAMS in kwargs.keys() and callable(sql_command):
                # 取出这个特殊参数，并从参数字典中移除
                decorator_sql_params = kwargs.pop(DECORATOR_SQL_PARAMS)
                new_command = sql_command(*decorator_sql_params)
            # 如果不包含'decorator_sql_params' 参数，且sql_command是字符类型
            elif DECORATOR_SQL_PARAMS not in kwargs and type(sql_command) == str:
                new_command = __convert_entity_name(sql_command=sql_command, entity=self.entity)
            else:
                raise ValueError('sql装饰器的使用方式错误，请查阅使用文档')
            self.cursor.execute(new_command, __combine_args_and_kwargs(args, kwargs))
            return self.cursor.fetchall()

        return wrapper

    return decorator


def Insert(func):  # noqa
    """插入装饰器，无需外部实现sql"""

    def wrapper(self, entity):
        """entity参数是传入的数据类实例
        而self.entity的数据类
        """

        # 获取entity类的属性名和类型
        fields = [field_name for field_name, _ in self.entity.__annotations__.items()]

        # 过滤掉有自增主键约束的属性
        fields = [field_name for field_name in fields
                  if Constraint.auto_primary_key[0]
                  not in get_constraint_from_property(getattr(self.entity, field_name, None))]

        sql_command = f"insert into {self.entity.__name__} " \
                      f"({', '.join(field_name for field_name in fields)}) " \
                      f"values ({', '.join('?' for _ in fields)});"
        # 判断传入的参数是否是数据类的实例
        if isinstance(entity, self.entity):
            values = [getattr(entity, field_name) for field_name in fields]
        else:
            # 如果不是，抛出一个类型错误异常
            raise TypeError(f"传入参数类型和Entity不一致，应传入 {self.entity.__name__} 实例")

        self.cursor.execute(sql_command, values)
        return self.cursor.lastrowid

    return wrapper


# ====================================================================================================================
# 新增方法
def update_cursor(self, cursor):
    """更新dao中的游标"""
    self.cursor = cursor


def get_sql_create_table(self):
    """自动实现的建表语句"""
    table_name = self.entity.__name__.lower()
    # 获取字段名称和类型的字典
    fields = self.entity.__annotations__

    field_definitions = []
    foreign_key_constraints = []

    for field_name, field_type in fields.items():
        # 获取字段的SQL类型
        sql_type = __get_sql_type(field_type)
        # 获取字段的约束条件
        constraints = get_constraint_from_property(property_value=getattr(self.entity, field_name, None))

        foreign_key_constraint = [item for item in constraints
                                  if isinstance(item, tuple)]
        constraints = [item for item in constraints
                       if item not in foreign_key_constraint]
        # 有外键
        if len(foreign_key_constraint) == 1:
            foreign_key_constraint = __get_foreign_key(field_name=field_name,
                                                       constraint=foreign_key_constraint[0])
            foreign_key_constraints.append(foreign_key_constraint)

        elif len(foreign_key_constraint) > 1:
            raise TypeError('一个属性只能指定一个外键')

        # 合并其他约束条件
        constraint = " ".join(constraints)

        # 拼接字段定义
        field_definition = f"{field_name} {sql_type} {constraint}"
        field_definitions.append(field_definition)
    # 将外键约束列表添加到字段定义列表的末尾
    field_definitions.extend(foreign_key_constraints)
    # 默认为严格的数据类型约束
    sql_create_table = f"CREATE TABLE IF NOT EXISTS {table_name} ({', '.join(field_definitions)});"
    return sql_create_table


@Insert
def insert(self, entity) -> int:  # noqa
    pass


# ====================================================================================================================
# 模块内使用的方法
def __get_sql_type(python_type):
    """转换python注释类型为sql类型"""
    if python_type in {int, bool}:
        return "INTEGER"
    elif python_type == str:
        return "TEXT"
    elif python_type == float:
        return "REAL"
    elif python_type == bytes:
        return "BLOB"

    raise ValueError(f"ysql不支持该python数据类型: {python_type}")  # noqa


def __get_foreign_key(constraint: tuple, field_name: str):
    """生成外键约束的sql语句"""
    foreign_key_entity, foreign_key_name, foreign_key_delete_link, foreign_key_update_link = constraint

    if foreign_key_delete_link is None and foreign_key_update_link is None:
        return f"FOREIGN KEY ({field_name}) " \
               f"REFERENCES {foreign_key_entity}({foreign_key_name})"

    # 存在表关联的情况
    elif foreign_key_delete_link is not None and foreign_key_update_link is None:
        return f"FOREIGN KEY ({field_name}) " \
               f"REFERENCES {foreign_key_entity}({foreign_key_name}) ON DELETE {foreign_key_delete_link}"
    elif foreign_key_delete_link is None and foreign_key_update_link is not None:
        return f"FOREIGN KEY ({field_name}) " \
               f"REFERENCES {foreign_key_entity}({foreign_key_name}) ON UPDATE {foreign_key_update_link}"
    elif foreign_key_delete_link is not None and foreign_key_update_link is not None:
        return f"FOREIGN KEY ({field_name}) " \
               f"REFERENCES {foreign_key_entity}({foreign_key_name}) " \
               f"ON DELETE {foreign_key_delete_link} ON UPDATE {foreign_key_update_link} "


def __combine_args_and_kwargs(*args):  # 不能传入**kwargs，这由调用条件决定了
    """一维展开不定参数，返回列表形式"""

    def flatten_args(arg):
        if isinstance(arg, (list, tuple)):
            # 如果是列表或元组，递归展开每个元素
            return [item for sublist in map(flatten_args, arg) for item in sublist]
        elif isinstance(arg, dict):
            # 如果是字典，取出所有值并递归展开
            return flatten_args(list(arg.values()))
        else:
            # 否则返回单个值
            return [arg]

    args_list = flatten_args(args)
    return tuple(args_list)


def __convert_entity_name(sql_command: str, entity):
    """将表名代替符替换为真正的表名"""
    if sql_command.count(TABLE_SUBSTITUTE) > 1:
        raise TypeError(f"sql语句非法使用了表名代替符：{TABLE_SUBSTITUTE}")

    return sql_command.replace(TABLE_SUBSTITUTE, entity.__name__.upper())



# -*- coding: utf-8 -*-
import os
from dataclasses import dataclass

from ysql import Dao, Sql
from ysql import Path, MetaDatabase
from ysql import Entity, Constraint


# ====================================================================================================================
# 项目gitee地址：https://gitee.com/darlingxyz/ysql（请于此查看项目详细信息，如README文件）
# 如果直接下载源码使用，请将ysql文件夹放置在根目录中，与venv目录同级，否则需要重新导入其中各文件的依赖项
# 通过pip安装时，可以在任何py文件中运行以下代码来生成此示例文件
# from ysql import get_example
# get_example()
# ====================================================================================================================
# 结构说明：
# Entity定义数据表
# Dao定义数据表的操作接口，提供表级控制
# Database集成Dao便于统一对外，提供库级控制


# ====================================================================================================================
# 数据表类的定义形式
@Entity  # 使用Entity装饰器
@dataclass  # 指定为原生的dataclass数据类
class Student:
    score: float
    picture: bytes
    name: str = '好家伙', Constraint.not_null  # 同时设置默认值和约束，以逗号分开即可
    student_id: int = Constraint.auto_primary_key  # Constraint类提供多种字段约束，请查看源码了解


# ====================================================================================================================
# Dao类的定义形式
@Dao(Student)  # 通过Dao装饰器传入对应的数据类
class DaoStudent:
    # 已内置insert(entity)方法，可直接调用DaoStudent().insert(entity)，参数entity应为对应数据类的实例

    @Sql("select name, score, picture from student where student_id=?;")  # 通过Sql装饰器传递标准sql语句
    def get_student(self, student_id):
        pass  # 装饰器会自动实现该方法
        # 使用select的时候返回结果为list[named_tuple]，其中每条记录会自动解析为具名元组。


# ====================================================================================================================
# 数据库类的定义形式
db_path = "db_folder/test_simple.db"
@Path(db_path)  # 需通过Path装饰器传递数据库的路径
class MyDatabase(MetaDatabase):  # 需继承MetaDatabase类，请查看其源码了解可调用的方法
    dao_student = DaoStudent()  # 需将Dao类写为Database类的静态类变量，以便集成Dao到一个数据类中对外访问


if __name__ == '__main__':
    if os.path.exists(db_path):
        os.remove(db_path)

    db = MyDatabase()  # 创建数据库类的实例
    db.connect()  # 连接数据库
    db.create_tables()  # 创建表

    db.dao_student.insert(Student(name="好家伙", score=95.6, picture=b'1235'))  # 插入一条student记录
    db.commit()  # 提交改动

    result = db.dao_student.get_student(student_id=1)  # 获取一条记录
    # 查询结果以列表形式返回
    print(f"查询结果：{result}")

    db.disconnect()  # 关闭数据库连接



# -*- coding: utf-8 -*-

import os
import sqlite3

from collections import namedtuple


# 传递路数据库路径参数的装饰器
def Path(path: str):
    def wrapper(cls):
        # 确保目录存在
        db_folder = os.path.dirname(path)
        if db_folder != '' and not os.path.exists(db_folder):
            os.makedirs(db_folder, exist_ok=True)
        # 新增数据库路径属性
        setattr(cls, "db_path", path)
        return cls

    return wrapper


class MetaDatabase:
    """数据库类，提供库级的操作"""

    # ================================================================================================================
    # 对外提供的可调用方法
    def connect(self, if_use_multithreading=True):
        """连接数据库"""
        # 设置多线程连接
        self.connection = sqlite3.connect(self.db_path, check_same_thread=not if_use_multithreading)  # noqa
        # 修改row_factory，使得返回结果为具名元组
        self.connection.row_factory = self.dict_factory
        self.cursor = self.connection.cursor()
        self.__update_cursor()

    def disconnect(self):
        """断开数据库连接"""
        if self.connection:
            self.connection.close()

    def create_tables(self):
        """创建全部表"""
        for dao in self.__dao_list:
            sql_command = dao.get_sql_create_table()
            self.cursor.execute(sql_command)
        self.commit()

    def commit(self):
        """提交事务"""
        self.connection.commit()

    def rollback(self):
        """回滚事务"""
        self.connection.rollback()

    def turn_on_foreign_key(self):
        """开启外键"""
        self.execute('PRAGMA foreign_keys=ON;')

    def execute(self, command: str):
        """给外部提供的直接执行sql的接口，避免了再调用connection"""
        self.connection.execute(command)

    # ================================================================================================================
    # 非外部调用方法
    def __new__(cls, *args, **kwargs):
        # 获取子类的所有属性
        subclass_attrs = dir(cls)

        # 初步筛选静态变量（不包括方法和特殊属性）
        static_attrs = [attr for attr in subclass_attrs
                        if not callable(getattr(cls, attr)) and not attr.startswith("__")]
        # 根据是否具有entity属性筛选出最终dao属性
        dao_list = [getattr(cls, attr) for attr in static_attrs
                    if hasattr(getattr(cls, attr), "entity")
                    and hasattr(getattr(cls, attr), "update_cursor")]

        # 赋值dao_list为类属性，以便在类内部可见
        cls.__dao_list = dao_list

        # 暂不支持单例模式
        return super().__new__(cls, *args, **kwargs)

    def __init__(self):
        self.connection = None
        self.cursor = None

    # ================================================================================================================
    # 类内使用的方法
    def __update_cursor(self):
        for dao in self.__dao_list:
            dao.update_cursor(cursor=self.cursor)

    @staticmethod
    def dict_factory(cursor, row: tuple):
        """转换查询结果格式为具名元组"""
        # 获取列描述信息
        fields = [column[0] for column in cursor.description]
        cls = namedtuple("Record", fields)
        return cls._make(row)  # noqa



# -*- coding: utf-8 -*-

import os
import shutil


def get_example():
    """复制本文件的工具函数"""
    # 获取当前文件的路径
    src = os.path.join(os.path.dirname(__file__), 'example_simple.py')
    dst = "ysql_example_simple.py"
    shutil.copy(src, dst)
    print('已生成示例文件：ysql_example_simple.py')

    src = os.path.join(os.path.dirname(__file__), 'example_detailed.py')
    dst = "ysql_example_detailed.py"
    shutil.copy(src, dst)
    print('已生成示例文件：ysql_example_detailed.py')



# -*- coding: utf-8 -*-

from typing import Union


def Entity(cls):
    """更改返回的值，如果不使用此装饰器，则始终完整返回默认值（包括约束），这在单独使用数据类的时候很不便，但对于建表（需要知道约束）和插入（需要知道主键约束）不影响
    仅在单独使用数据类的时候有效"""
    orig_getattribute = cls.__getattribute__

    def new_getattribute(self, name):
        value = orig_getattribute(self, name)
        return get_value_from_property(value)

    cls.__getattribute__ = new_getattribute
    return cls


class Constant:
    """存储特殊形式的常量"""
    PRIMARY_KEY = ("PRIMARY KEY",)
    AUTO_PRIMARY_KEY = ("PRIMARY KEY AUTOINCREMENT",)
    NOT_NULL = ("NOT NULL",)
    UNIQUE = ("UNIQUE",)

    NO_ACTION = 'NO ACTION'
    RESTRICT = 'RESTRICT'
    CASCADE = 'CASCADE'
    SET_NULL = 'SET NULL'
    SET_DEFAULT = 'SET DEFAULT'


class Constraint:
    """提供对字段的各种约束"""

    # =================================================================================================================
    # 1.可直接使用的约束常量
    primary_key = Constant.PRIMARY_KEY  # 主键
    auto_primary_key = Constant.AUTO_PRIMARY_KEY  # 自增主键
    not_null = Constant.NOT_NULL  # 非空
    unique = Constant.UNIQUE  # 唯一

    # 针对外键的约束
    no_action = Constant.NO_ACTION
    cascade = Constant.CASCADE
    set_null = Constant.SET_NULL
    restrict = Constant.RESTRICT
    set_default = Constant.SET_DEFAULT

    # =================================================================================================================
    # 2.需要外部传值的约束
    @staticmethod
    def default(default_value: Union[int, str, float, bytes]):
        """默认值约束"""
        if type(default_value) in {int, str, float, bytes}:
            return (f'DEFAULT {default_value}',)  # noqa
        raise TypeError(f'数据类型不匹配，允许的数据类型为：int, str, float, bytes')

    @staticmethod
    def check(check_condition: str):
        """条件约束"""
        if type(check_condition) in {str}:
            return (f'CHECK({check_condition})',)  # noqa
        raise TypeError(f'数据类型不匹配，允许的数据类型为：str')

    @staticmethod
    def foreign_key(entity, field, delete_link=None, update_link=None):
        """外键约束"""
        return ((entity.__name__, field, delete_link, update_link),)  # noqa

    @staticmethod
    def comment(comment: str):
        """字段注释"""
        if type(comment) in {str}:
            return (f'-- {comment}\n',)  # noqa
        raise TypeError(f'数据类型不匹配，允许的数据类型为：str')


def get_constraint_from_property(property_value):
    """从属性值中解析出约束条件"""
    if not isinstance(property_value, tuple):
        return []

    # 单元素元组
    if len(property_value) == 1:
        return list(property_value)

    # 多元素元组
    elif len(property_value) >= 1:
        constraints = []

        for item in property_value:

            if not isinstance(item, tuple):
                continue
            # 元素也是元组，说明是约束条件
            if len(item) == 1:
                constraints.append(item[0])
            else:
                raise TypeError(f'约束条件必须为长度为1的元组，错误的数据：{item}')

        return constraints


def get_value_from_property(property_value):
    """从属性值中解析出是基本类型的唯一属性默认值"""
    # 属性值就是基本类型
    if isinstance(property_value, (int, str, float, bytes)) or property_value is None:
        return property_value

    # 检查是否为元组
    if not isinstance(property_value, tuple):
        raise TypeError('该属性值的类型不满足要求，只允许基本类型和元组')

    # 用列表保存匹配到的基本属性值，用来进一步检验该值是否唯一
    value_list = []
    for value in property_value:
        if isinstance(value, (int, str, float, bytes)):
            value_list.append(value)

    list_length = len(value_list)
    # 没有默认值
    if list_length == 0:
        return None
    # 有唯一的默认值
    elif list_length == 1:
        return value_list[0]
    else:
        raise TypeError('错误传递了多个基本数据类型的默认值，只允许有一个默认值')
